class Search(object):
    def __init__(self, search_list, data_iterator, binary_search=False):
        self._search_list = search_list
        self._data_iterator = data_iterator
        self._binary_search = binary_search
        if binary_search:
            self.convert_to_binary_search_list()
    
    def __iter__(self):
        return self

    def __next__(self):
        for row in self._data_iterator:
            match = self.search(row)
            if match:
                yield (row, match)

    def convert_to_binary_search_list(self):
        new_list = {}
        for key in self._search_list:
            # stores both new binary key and the old
            new_list[bytes(key,'utf-8')] = self._search_list[key]
            new_list[key] = self._search_list[key]
        self._search_list = new_list

    def maybe_convert_to_binary(self, value):
        if type(value) != str:
            return value
        if self._binary_search:
            return bytes(value, 'utf-8')
        return value

    def maybe_convert_binary_list(self, values):
        if not self._binary_search:
            return values

        new_list = []
        for value in values:
            new_list.append(self.maybe_convert_to_binary(value))
        return new_list




#!/usr/bin/python3

"""This is a generic threat-feed searching application that takes a
threat feed of various types and matches it against sources of network
log data to search through.  It can take both threat-feeds and log
data from multiple sources, including kafka and files.  It's output is
configurable and can be given a jinja2 template for generic report
generation.
"""

import datetime
import yaml

import sys
import time
import argparse
from msgpack import unpackb

# ick, remote during packaging:
import sys, os

from gawseed.threatfeed.feeds.kafka import KafkaThreatFeed
from gawseed.threatfeed.feeds.fsdb import FsdbThreatFeed

from gawseed.threatfeed.datasources.kafka import KafkaDataSource
from gawseed.threatfeed.datasources.fsdb import FsdbDataSource
from gawseed.threatfeed.datasources.bro import BroDataSource

from gawseed.threatfeed.search.ip import IPSearch
from gawseed.threatfeed.search.http import HTTPSearch
from gawseed.threatfeed.search.ssh import SSHSearch

from gawseed.threatfeed.events.printer import EventStreamPrinter
from gawseed.threatfeed.events.dumper import EventStreamDumper
from gawseed.threatfeed.events.reporter import EventStreamReporter

debug = False

def parse_args():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter, description="This uses data from threat feeds (in kafka topics or files) to search through network data (from kafka or files) and reports any matches as 'events'.")

    # THREAT FEEDS
    # ----------------------------------------------------------------------
    # arguments for the threat feed source
    group = parser.add_argument_group("Threat feed arguments")
    group.add_argument("-B", "--threat-begin-time", default=None, type=str,
                        help="Date to start returning threat feed data from")

    group.add_argument("-E", "--threat-end-time", default=None, type=str,
                        help="Date to stop returning threat feed data after")

    group.add_argument("-M", "--threat-max-records", default=None, type=int,
                        help="Maximum number of threat records to retrieve and search for ")

    group.add_argument("--threat-timeout", default=1000, type=int,
                        help="Maximum number of ms to wait for a threat feed before figuring its done")


    # arguments for kafka threat feeds
    group = parser.add_argument_group("Kafka specific threat feed arguments")
    group.add_argument("-K", "--threat-kafka-servers", default=[], type=str,
                        nargs="*", help="Kafka server to pull topics from")

    group.add_argument("--threat-kafka-topic", default="gawseed_ipaddresses", type=str,
                        help="The kafka threat topic to use as the threat source")

    group = parser.add_argument_group("File/FSDB specific threat feed arguments")
    # arguments for file based threat feeds
    group.add_argument("--threat-fsdb", default=None, type=str,
                        help="FSDB based threat feed")

    # DATA FEEDS
    # ----------------------------------------------------------------------
    # arguments for the data feeds
    group = parser.add_argument_group("Data feed arguments")
    group.add_argument("-T", "--time-column", default="ts", type=str,
                        help="The time column to read looking for end-time")

    group.add_argument("-m", "--max-records", type=int,
                        help="maximum number of records to return")

    group.add_argument("-b", "--begin-time", default=None, type=str,
                        help="Date to start returning threat feed data from")

    group.add_argument("-e", "--end-time", default=None, type=str,
                        help="Date to stop returning threat feed data after")

    # kafka specific data feed options
    group = parser.add_argument_group("Kafka specific data feed arguments")
    group.add_argument("--data-kafka-servers", default=[], type=str,
                        nargs="*", help="Kafka server to pull data streams from")

    group.add_argument("-t", "--data-topic", default="ssh", type=str,
                        help="Kafka topic to request from the data source")

    group = parser.add_argument_group("BRO specific data feed arguments")
    # BRO file
    group.add_argument("--bro-data", type=str,
                        help="Bro file to read data from")

    group = parser.add_argument_group("File/FSDB specific data feed arguments")
    # fsdb data specific options
    group.add_argument("--fsdb-data", default=None, type=str,
                        help="FSDB data file for testing")


    group = parser.add_argument_group("Output formatting arguments")
    # OUTPUT
    # ----------------------------------------------------------------------
    # output processing
    group.add_argument("-f", "--fields", default=None, type=str, nargs="*",
                        help="Fields to print")
    
    group.add_argument("--fsdb", action="store_true",
                        help="Output FSDB formatted data")

    group.add_argument("-o", "--output-pattern", default=None, type=str,
                        help="Output a pattern of files instead of stdout; include %%d for a count")

    group.add_argument("-j", "--jinja-template", default=None, type=argparse.FileType("r"),
                        help="The jinja template to use when generating reports")

    group.add_argument("-J", "--jinja-extra-information", default=None, type=argparse.FileType("r"),
                        help="Extra information in YAML format to include with report generation in 'extra' an field")

    # Configuration
    group = parser.add_argument_group("Global configuration")
    group.add_argument("-y", "--config", type=argparse.FileType("r"),
                       help="A YAML configuration file specifying all modules to be loaded")

    # DEBUGGING
    group = parser.add_argument_group("Debugging arguments")
    group.add_argument("--dump-threat-feed", action="store_true",
                        help="Simply dump the threat feed data to stdout")
   
    group.add_argument("--dump-data", action="store_true",
                        help="Simply dump all the data without searching to stdout")

    group.add_argument("--dump-events", action="store_true",
                        help="Simply dump all the event data in raw form")

    group.add_argument("-V", "--verbose", action="store_true",
                        help="Verbose/Debugging output")

    args = parser.parse_args()
    if args.verbose:
        global debug
        debug = True

    # if args.merge_grep and not args.merge_key:
    #     raise ValueError("--merge-key/-k is required with --merge-grep")

    return args

def verbose(msg):
    if debug:
        print(msg)

def load_yaml_config(args):
    defaults = {
        'threatsource': {
            'type': 'kafka',
            'timeout': 2000,
        },
        'datasource': {
            'type': 'kafka',
            'topic': 'ssh', # XXX don't make this a default?
        },
        'searcher': {
            'module': 'ssh' # XXX same
        },
        'reporter': {
            'module': 'eventReporter',
        }
    }

    conf = yaml.load(args.config, Loader=yaml.FullLoader)
    
    # fill in defaults
    threatconfs = conf['threat-search']
    for threatconf in threatconfs:
        # insert default values if needed
        for key in defaults:
            if key not in threatconf:
                threatconf[key] = {}
            for subkey in defaults[key]:
                if subkey not in threatconf[key]:
                    threatconf[key][subkey] = defaults[key][subkey]

    # load modules?
    module_xforms = {
        'datasource': {
            'ssh': 'gawseed.threatfeed...'
        }
    }



def get_threat_feed(args):
    """Read in the threat feed stream as a data source to search for"""

    # read in the threat feed stream as a data source to search for
    if args.threat_fsdb:
        threat_source = FsdbThreatFeed(args.threat_fsdb)
    else:
        threat_source = KafkaThreatFeed(args.threat_kafka_servers, args.threat_begin_time,
                                        args.threat_kafka_topic, timeout=args.threat_timeout)
    verbose("created threat feed: " + str(threat_source))

    threat_source.open()
    (search_data, search_index) = threat_source.read(args.threat_max_records)
    verbose("  read feed with " + str(len(search_data)) + " entries")

    if args.dump_threat_feed:
        print(search_data)
        print(search_index)
        exit(0)

    return (threat_source, search_data, search_index)


def get_data_source(args):
    """Get the data source and open it for traversing"""
    if args.fsdb_data:
        data_source = FsdbDataSource(file=args.fsdb_data)
    elif args.bro_data:
        data_source = BroDataSource(file=args.bro_data)
    else:
        data_source = KafkaDataSource(args.data_kafka_servers,
                                      begin_time = args.begin_time,
                                      topic=args.data_topic)

    verbose("created data feed: " + str(data_source))
    data_source.open()

    # just print it?
    if args.dump_data:
        for finding in data_source:
            print(finding)
        exit(0)

    return data_source

def get_searcher(args, search_index, data_source):
    """Create a searcher object"""
    # create the searching interface
    if args.data_topic == 'ssh':
        searcher = SSHSearch(search_index, data_iterator = data_source, binary_search = data_source.is_binary())
    elif args.data_topic == 'ip' or args.data_topic == 'conn':
        searcher = IPSearch(search_index, data_iterator = data_source, binary_search = data_source.is_binary())
    elif args.data_topic == 'http':
        searcher = HTTPSearch(search_index, data_iterator = data_source, binary_search =data_source.is_binary())
        
    verbose("created searcher: " + str(searcher))

    return searcher

def load_modules_from_args(args):

def main():
    args = parse_args()

    (threat_source, search_data, search_index) = get_threat_feed(args)
    data_source = get_data_source(args)
    searcher = get_searcher(args, search_index, data_source)

    #output = EventStreamDumper() 
    if args.dump_events:
        output = EventStreamDumper(stream=args.output_pattern)
    elif args.jinja_template:
        output = EventStreamReporter(stream=args.output_pattern,
                                     jinja_template=args.jinja_template,
                                     jinja_extra_information=args.jinja_extra_information)
    else:
        output = EventStreamPrinter(stream=args.output_pattern,
                                    extra_fields=['auth_success']) # auth for ssh

    verbose("created output: " + str(output))

    # loop through all the data for matches
    for count, finding in enumerate(next(searcher)):
        output.new_output(count)
        output.write(count, finding[0], finding[1])
        output.maybe_close_output()
        if args.max_records and count >= args.max_records:
            break

if __name__ == "__main__":
    main()



from dateutil import parser
import json

from kafka import KafkaConsumer
from kafka.structs import TopicPartition

class KafkaThreatFeed():
    """Pulls threat data from a GAWSEED project (or other) kafka threat-feed source."""
    def __init__(self, bootstrap_servers, begin_time=None, topic="gawseed_ipaddresses", partition=0, timeout=1000):
        self._bootstrap_servers = bootstrap_servers
        self._topic = topic
        self._partition = partition
        self._begin_time = begin_time
        self._timeout=timeout

    def open(self):
        self._consumer = KafkaConsumer(bootstrap_servers=self._bootstrap_servers,
                                       consumer_timeout_ms=self._timeout)

        # point to what we want at
        partition = TopicPartition(self._topic, self._partition)
        self._consumer.assign([partition])

        offset = None
        if self._begin_time:
            timestamp = parser.parse(self._begin_time).timestamp() * 1000
            offinfo = self._consumer.offsets_for_times({partition: timestamp})
            if offinfo == None or offinfo[partition] == None:
                raise ValueError("There is no data in the threat feed stream the begin date")
            offset = offinfo[partition].offset
            self._consumer.seek(partition, offset)

    def __iter__(self):
        return self

    def parse_record(self, record):
        entry = json.loads(record.value)
        return entry

    def __next__(self):
        return next(self._consumer)

    def read(self, max_records=None, value_column='value'):
        array = []
        dictionary = {}

        if self._begin_time:
            timestamp = parser.parse(self._begin_time).timestamp()

        for (count, entry) in enumerate(self._consumer):
            entry = self.parse_record(entry)

            # tmp hack to work around kafka hanging on some topics;
            # thus we start from the beginning and read everything.
            # this naturally won't scale.
            if timestamp and int(entry['timestamp']) < timestamp:
                continue

            array.append(entry)
            dictionary[entry[value_column]] = entry # note, may erase older ones; build array?
            if max_records and count >= max_records:
                break

        return (array, dictionary)




from gawseed.threatfeed.search import Search

class IPSearch(Search):
    def __init__(self, search_list,
                 data_iterator,
                 binary_search=False,
                 search_keys = ['id_orig_h', 'id_resp_h']):
        super().__init__(search_list, data_iterator, binary_search)
        self._search_keys = self.maybe_convert_binary_list(search_keys)
    
    def search(self, row):
        for key in self._search_keys:
            if row[key] in self._search_list:
                return self._search_list[row[key]]
        return None




import sys

class EventStream():
    def __init__(self, stream=sys.stdout):
        if type(stream) == str:
            self._stream_pattern = stream
        else:
            self._stream_pattern = None
            if stream == None:
                self._stream = sys.stdout
            else:
                self._stream = stream

    def new_output(self, count):
        if self._stream_pattern:
            filename = self._stream_pattern % (count)
            self._stream = open(filename, "w")

    def maybe_close_output(self):
        if self._stream_pattern:
            self._stream.close()

    def output(self, something):
        self._stream.write(something)

    def maybe_convert(self, row):
        new_row = {}
        for key in row:
            value = row[key]
            if type(key) == bytes:
                key = key.decode()
            if type(value) == bytes:
                new_row[key] = value.decode()
            else:
                new_row[key] = value
        return new_row

    def write(self, count, row, match):
        row = self.maybe_convert(row)
        self.write_row(count, row, match)




import pyfsdb

class FsdbThreatFeed():
    def __init__(self, fsdb_file=None, value_column="value"):
        self._fsdb_file = fsdb_file
        self._value_column = value_column

    def open(self):
        self._tfh = pyfsdb.Fsdb(self._fsdb_file,
                                return_type=pyfsdb.RETURN_AS_DICTIONARY)
        self._tfh_index_column = self._tfh.get_column_number(self._value_column)

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._tfh)

    def column_names(self):
        return self._tfh.column_names

    def index_column_number(self):
        return self._tfh.get_column_number(self._value_column)

    def read(self, max_records = None):
        array = []
        dictionary = {}
        for (count,entry) in enumerate(self._tfh):
            array.append(entry)
            dictionary[entry[self._value_column]] = entry # note, may erase older ones; build array?
            if max_records and count+1 >= max_records:
                break

        return (array, dictionary)

        



from dateutil import parser

from kafka import KafkaConsumer
from kafka.structs import TopicPartition
from msgpack import unpackb

from . import DataSource

class KafkaDataSource(DataSource):
    def __init__(self, bootstrap_servers,
                 begin_time=None, topic="ssh", max_records=None,
                 consumer_timeout_ms=None):
        super().__init__()
        self._bootstrap_servers = bootstrap_servers
        self._begin_time = begin_time
        self._topic = topic
        self._max_records = max_records
        self._consumer_timeout_ms = consumer_timeout_ms

    def open(self):
        consumer = KafkaConsumer(bootstrap_servers=self._bootstrap_servers)
                                 #consumer_timeout_ms=self._consumer_timeout_ms)
        # point to what we want at
        partition = TopicPartition(self._topic,0)
        consumer.assign([partition])

        if self._begin_time:
            timestamp = parser.parse(self._begin_time).timestamp() * 1000
            offinfo = consumer.offsets_for_times({partition: timestamp})
            if offinfo == None or offinfo[partition] == None:
                raise ValueError("There is no data in the enterprise stream the begin date")
            offset = offinfo[partition].offset
            consumer.seek(partition, offset)

        self._consumer = consumer
        return self

    def __iter__(self):
        return self

    def __next__(self):
        row = next(self._consumer)
        decoded_row = unpackb(row.value)
        return decoded_row

    def is_binary(self):
        return True





from pyfsdb.Fsdb import RETURN_AS_DICTIONARY,Fsdb

from gawseed.threatfeed.datasources.fsdb import FsdbDataSource

class BroDataSource(FsdbDataSource):
    def __init__(self, file_handle=None, file=None):
        self._file_handle = file_handle
        self._file = file

    def open(self):
        if self._file and not self._file_handle:
            self._file_handle = open(self._file, "r")
        
        column_names = None
        for line in self._file_handle:
            if line[0] != "#": # skip at end of headers
                break

            if line[0:7] == "#fields":
                column_names = line.replace(".", "_").split("\t")
                column_names.pop(0)

        if not column_names:
            raise ValueError("passed bro file isn't in expected bro file format")

        self._fh = Fsdb(file_handle=self._file_handle,
                        return_type=RETURN_AS_DICTIONARY)
        self._fh.column_names = column_names
        # todo:: XXX: we eat a line here; fix this



import sys

from gawseed.threatfeed.events import EventStream

class EventStreamPrinter(EventStream):
    def __init__(self, stream=sys.stdout, extra_fields=[]):
        super().__init__(stream)
        self._form = "  %-30.30s: %s\n"
        self._extra_fields = extra_fields

    def out(self, info1, info2):
        self.output(self._form % (info1 + ":", info2))

    def write_row(self, count, row, match):
        self.output("match #" + str(count) + ":\n")

        self.output("Threat information:\n")
        self.out("Match value", match['value'])
        self.out("Tag", match['tag'])

        self.output("Data source information:\n")
        self.out("Timestamp", row['ts'])
        self.out("id_orig_h", row['id_orig_h'])
        self.out("id_resp_h", row['id_resp_h'])

        if len(self._extra_fields) > 0:
            self.output("Extra information:\n")
            for field in self._extra_fields:
                if field in row:
                    self.out(field, row[field])
                if field in match:
                    self.out(field, match[field])

        self.output("-" * 70 + "\n")




import sys

from gawseed.threatfeed.events import EventStream

class EventStreamDumper(EventStream):
    def __init__(self, stream=sys.stdout):
        super().__init__(stream)

    def write(self, count, row, match):
        self.output("match #" + str(count) + ":\n")
        self.output(str(row) + "\n")
        self.output("--\n")
        self.output(str(match) + "\n")
        self.output("--------------")




import sys
import re

from gawseed.threatfeed.search import Search

class HTTPSearch(Search):
    def __init__(self, search_list,
                 data_iterator=None,
                 binary_search=False):
        super().__init__(search_list, data_iterator, binary_search)

        self._relist = []
        for item in search_list:
            try: 
                compiled = re.compile(item)
                self._relist.append({ 'match': search_list[item],
                                      're': compiled})
            except:
                sys.stderr.write("failed to compile pattern: %s" % (item))
    
    def search(self, row):
        if b'host' not in row or b'uri' not in row or row[b'host'] is None or row[b'uri'] is None:
            return None
        if self._binary_search:
            url = "http://" + row[b'host'].decode() + "/" + row[b'uri'].decode()
        else:
            url = "http://" + row[b'host'].decode() + "/" + row[b'uri'].decode()

        for expr in self._relist:
            match = expr['re'].match(url)
            if match:
                return expr['match']




import pyfsdb
from . import DataSource

class FsdbDataSource(DataSource):
    def __init__(self, file_handle=None, file=None):
        self._file_handle = file_handle
        self._file = file

    def open(self):
        self._fh = fsdb.Fsdb(file_handle=self._file_handle, filename=self._file,
                             return_type=fsdb.RETURN_AS_DICTIONARY)

    def __iter__(self):
        return self

    def __next__(self):
        return (next(self._fh))




import sys
import jinja2

from gawseed.threatfeed.events import EventStream

class EventStreamReporter(EventStream):
    def __init__(self, stream=sys.stdout, jinja_template=None, jinja_extra_information=None):
        super().__init__(stream)
        self._jinja_template = jinja_template.read()
        if jinja_extra_information:
            self._jinja_extra_information = yaml.load(jinja_extra_information)
        else:
            self._jinja_extra_information = None

        self._template = jinja2.Template(self._jinja_template)

    def write_row(self, count, row, match):
        output = self._template.render({ 'count': count,
                                         'row': row,
                                         'match': match,
                                         'extra': self._jinja_extra_information})
        self.output(output)




from gawseed.threatfeed.search.ip import IPSearch

class SSHSearch(IPSearch):
    def __init__(self, search_list,
                 data_iterator=None,
                 binary_search=False,
                 search_keys=['id_orig_h', 'id_resp_h'],
                 auth_success_key='auth_success',
                 auth_success_value=True):
        super().__init__(search_list, data_iterator, binary_search, search_keys)
        # Really need a tri-nary option for this
        self._auth_success_key = self.maybe_convert_to_binary(auth_success_key)
        self._auth_success_value = self.maybe_convert_to_binary(auth_success_value)
    
    def search(self, row):
        if not self._auth_success_key or self._auth_success_key not in row:
            return None
        if row[self._auth_success_key] != self._auth_success_value:
            return None
        return super().search(row)

